@@ -1,19 +1,20 @@
 from .node_graph import NodeGraph
 from .node import Node
 from .decorator import node
-from .executor import NodeExecutor
+from .executor import SafeExecutor, RuntimeExecutor
 from .nodes import NodePool
 from .collection import group
 from .socket_spec import namespace, dynamic
 
-__version__ = "0.3.5"
+__version__ = "0.4.0"
 
 
 __all__ = [
     "NodeGraph",
     "Node",
     "node",
-    "NodeExecutor",
+    "SafeExecutor",
+    "RuntimeExecutor",
     "NodePool",
     "group",
     "namespace",
