@@ -3,18 +3,59 @@ from node_graph.registry import RegistryHub, registry_hub
 from node_graph.collection import NodeCollection, LinkCollection
 from uuid import uuid1
 from node_graph.node_spec import NodeSpec
-from node_graph.socket_spec import SocketSpec, BaseSocketSpecAPI
+from node_graph.socket_spec import SocketSpec, SocketSpecAPI
 from typing import Dict, Any, List, Optional, Union, Callable
 import yaml
 from node_graph.node import Node
 from node_graph.socket import NodeSocket
 from node_graph.link import NodeLink
 from node_graph.utils import yaml_to_dict
-from node_graph_widget import NodeGraphWidget
-from .config import BuiltinPolicy, BUILTIN_NODES
+from .config import BuiltinPolicy, BUILTIN_NODES, MAX_LINK_LIMIT
+from .mixins import IOOwnerMixin, WidgetRenderableMixin
+from dataclasses import dataclass
+from dataclasses import replace
 
 
-class NodeGraph:
+@dataclass(frozen=True)
+class GraphSpec:
+    """Specification for a NodeGraph's inputs and outputs."""
+
+    schema_source: str = "EMBEDDED"
+    inputs: Optional[SocketSpec] = None
+    outputs: Optional[SocketSpec] = None
+    ctx: Optional[SocketSpec] = None
+
+    def __post_init__(self):
+        # ctx should be dynamic
+        if self.ctx is not None and not self.ctx.dynamic:
+            object.__setattr__(self, "ctx", replace(self.ctx, dynamic=True))
+
+    def to_dict(self) -> Dict[str, Any]:
+        data: Dict[str, Any] = {"schema_source": self.schema_source}
+        if self.inputs is not None:
+            data["inputs"] = self.inputs.to_dict()
+        if self.outputs is not None:
+            data["outputs"] = self.outputs.to_dict()
+        if self.ctx is not None:
+            data["ctx"] = self.ctx.to_dict()
+        return data
+
+    @classmethod
+    def from_dict(cls, data: Dict[str, Any]) -> "GraphSpec":
+        from node_graph.socket_spec import SocketSpec
+
+        inputs = SocketSpec.from_dict(data["inputs"]) if "inputs" in data else None
+        outputs = SocketSpec.from_dict(data["outputs"]) if "outputs" in data else None
+        ctx = SocketSpec.from_dict(data["ctx"]) if "ctx" in data else None
+        return cls(
+            schema_source=data.get("schema_source", "EMBEDDED"),
+            inputs=inputs,
+            outputs=outputs,
+            ctx=ctx,
+        )
+
+
+class NodeGraph(IOOwnerMixin, WidgetRenderableMixin):
     """A collection of nodes and links.
 
     Attributes:
@@ -41,23 +82,22 @@ class NodeGraph:
         >>> ng.to_dict()
     """
 
-    registry: Optional[RegistryHub] = registry_hub
+    _REGISTRY: Optional[RegistryHub] = registry_hub
+    _BUILTINS_POLICY = BuiltinPolicy()
+    _SOCKET_SPEC_API = SocketSpecAPI
+
     platform: str = "node_graph"
-    _socket_spec = BaseSocketSpecAPI
-    Builtins: BuiltinPolicy = BuiltinPolicy()
 
     def __init__(
         self,
         name: str = "NodeGraph",
         inputs: Optional[SocketSpec | List[str]] = None,
         outputs: Optional[SocketSpec | List[str]] = None,
+        ctx: Optional[SocketSpec | List[str]] = None,
         uuid: Optional[str] = None,
         graph_type: str = "NORMAL",
         graph: Optional[NodeGraph] = None,
         parent: Optional[Node] = None,
-        state: str = "CREATED",
-        action: str = "NONE",
-        description: str = "",
         interactive_widget: bool = False,
         init_graph_level_nodes: bool = True,
     ) -> None:
@@ -74,55 +114,62 @@ class NodeGraph:
         self.graph_type = graph_type
         self.graph = graph
         self.parent = parent
-        self.NodePool = self.registry.node_pool
-        self.SocketPool = self.registry.socket_pool
-        self.PropertyPool = self.registry.property_pool
-        self.type_mapping = dict(self.registry.type_mapping)
-        self.nodes = NodeCollection(graph=self, pool=self.NodePool)
+        self.type_mapping = dict(self._REGISTRY.type_mapping)
+        self.type_promotions = set(self._REGISTRY.type_promotion)
+        self.nodes = NodeCollection(graph=self, pool=self._REGISTRY.node_pool)
         self.links = LinkCollection(self)
-        self.state = state
-        self.action = action
-        self.description = description
         self._widget = None
         self.interactive_widget = interactive_widget
         self._version = 0  # keep track the changes
-        self._inputs = inputs
-        self._outputs = outputs
+        self._init_graph_spec(inputs, outputs, ctx)
         if init_graph_level_nodes:
-            self._init_graph_level_nodes(inputs, outputs)
+            self._init_graph_level_nodes()
 
-    def _init_graph_level_nodes(self, inputs, outputs, ctx=None):
-        from dataclasses import replace
+        self.state = "CREATED"
+        self.action = "NONE"
+        self.description = ""
 
-        inputs = self._socket_spec.validate_socket_data(inputs)
-        outputs = self._socket_spec.validate_socket_data(outputs)
-        ctx = self._socket_spec.validate_socket_data(ctx)
+    def _init_graph_spec(
+        self,
+        inputs: Optional[SocketSpec],
+        outputs: Optional[SocketSpec],
+        ctx: Optional[SocketSpec],
+    ) -> None:
 
-        base_class = self.NodePool["graph_level"].load()
-
-        # graph inputs
-        inputs = self._socket_spec.dynamic(Any) if inputs is None else inputs
-        meta = replace(inputs.meta, sub_socket_default_link_limit=1000000)
+        inputs = self._SOCKET_SPEC_API.validate_socket_data(inputs)
+        # if inputs is None, we assume it's a dynamic inputs
+        inputs = self._SOCKET_SPEC_API.dynamic() if inputs is None else inputs
+        meta = replace(inputs.meta, child_default_link_limit=MAX_LINK_LIMIT)
         inputs = replace(inputs, meta=meta)
+        outputs = self._SOCKET_SPEC_API.validate_socket_data(outputs)
+        # if outputs is None, we assume it's a dynamic outputs
+        outputs = self._SOCKET_SPEC_API.dynamic() if outputs is None else outputs
+        ctx = self._SOCKET_SPEC_API.validate_socket_data(ctx)
+        # if ctx is None, we assume it's a dynamic ctx
+        ctx = self._SOCKET_SPEC_API.dynamic() if ctx is None else ctx
+        self.spec = GraphSpec(inputs=inputs, outputs=outputs, ctx=ctx)
+
+    def _init_graph_level_nodes(self):
+        base_class = self._REGISTRY.node_pool["graph_level"].load()
         self.graph_inputs_spec = NodeSpec(
             identifier="graph_inputs",
-            inputs=inputs,
+            inputs=self.spec.inputs,
             base_class=base_class,
         )
         self.nodes._new(self.graph_inputs_spec, name="graph_inputs")
-        # graph outputs
-        outputs = self._socket_spec.dynamic(Any) if outputs is None else outputs
+
         self.graph_outputs_spec = NodeSpec(
             identifier="graph_outputs",
-            inputs=outputs,
+            inputs=self.spec.outputs,
             base_class=base_class,
         )
         graph_outputs = self.nodes._new(self.graph_outputs_spec, name="graph_outputs")
         graph_outputs.inputs._name = "outputs"
         # graph context
-        ctx = self._socket_spec.dynamic(Any) if ctx is None else ctx
-        meta = replace(ctx.meta, sub_socket_default_link_limit=1000000)
+        ctx = self.spec.ctx or self._SOCKET_SPEC_API.dynamic(Any)
+        meta = replace(ctx.meta, child_default_link_limit=1000000)
         ctx = replace(ctx, meta=meta)
+        self.spec = replace(self.spec, ctx=ctx)
         self.graph_ctx_spec = NodeSpec(
             identifier="graph_ctx",
             inputs=ctx,
@@ -175,53 +222,60 @@ class NodeGraph:
     def ctx(self, value: Dict[str, Any]) -> None:
         """Set context node."""
         self.graph_ctx.inputs._clear()
-        self.graph_ctx.inputs._set_socket_value(value, link_limit=100000)
+        self.graph_ctx.inputs._set_socket_value(value)
 
-    def generate_inputs(self) -> None:
+    def update_ctx(self, value: Dict[str, Any]) -> None:
+        self.ctx._set_socket_value(value)
+
+    def expose_inputs(self, names: Optional[List[str]] = None) -> None:
         """Generate group inputs from nodes."""
-        self.inputs._clear()
-        for node in self.nodes:
-            if node.name in BUILTIN_NODES:
-                continue
-            # skip linked sockets
-            socket = node.inputs._copy(
-                node=self.graph_inputs,
-                parent=self.inputs,
-                skip_linked=True,
-                skip_builtin=True,
-            )
-            socket._name = node.name
-            self.inputs._append(socket)
-            keys = node.inputs._get_all_keys()
-            exist_keys = socket._get_all_keys()
-            for key in keys:
-                new_key = f"{node.name}.{key}"
-                if new_key not in exist_keys:
-                    continue
-                # add link from group inputs to node inputs
-                self.add_link(self.inputs[new_key], node.inputs[key])
+        from node_graph.socket_spec import remove_spec_field
 
-    def generate_outputs(self) -> None:
+        all_names = set(self.nodes._get_keys())
+        names = set(names or all_names)
+        missing = names - all_names
+        if missing:
+            raise ValueError(f"The following named nodes do not exist: {missing}")
+        for name in names - set(BUILTIN_NODES):
+            node = self.nodes[name]
+            # update the _inputs spec
+            if node.spec.inputs is not None:
+                # skip linked sockets
+                names = [
+                    link.to_socket._scoped_name
+                    for link in self.links
+                    if link.to_node == node
+                ]
+                spec = remove_spec_field(node.spec.inputs, names=names)
+                socket = self.add_input_spec(spec, name=node.name)
+                keys = node.inputs._get_all_keys()
+                exist_keys = socket._get_all_keys()
+                for key in keys:
+                    new_key = f"{node.name}.{key}"
+                    if new_key not in exist_keys:
+                        continue
+                    # add link from group inputs to node inputs
+                    self.add_link(self.inputs[new_key], node.inputs[key])
+
+    def expose_outputs(self, names: Optional[List[str]] = None) -> None:
         """Generate group outputs from nodes."""
-        self.outputs._clear()
-        for node in self.nodes:
-            if node.name in BUILTIN_NODES:
-                continue
-            socket = node.outputs._copy(
-                node=self.graph_outputs,
-                parent=self.outputs,
-                skip_builtin=True,
-            )
-            socket._name = node.name
-            self.outputs._append(socket)
-            keys = node.outputs._get_all_keys()
-            exist_keys = socket._get_all_keys()
-            for key in keys:
-                new_key = f"{node.name}.{key}"
-                if new_key not in exist_keys:
-                    continue
-                # add link from node outputs to group outputs
-                self.add_link(node.outputs[key], self.outputs[new_key])
+        all_names = set(self.nodes._get_keys())
+        names = set(names or all_names)
+        missing = names - all_names
+        if missing:
+            raise ValueError(f"The following named nodes do not exist: {missing}")
+        for name in names - set(BUILTIN_NODES):
+            node = self.nodes[name]
+            if node.spec.outputs is not None:
+                socket = self.add_output_spec(node.spec.outputs, name=node.name)
+                keys = node.outputs._get_all_keys()
+                exist_keys = socket._get_all_keys()
+                for key in keys:
+                    new_key = f"{node.name}.{key}"
+                    if new_key not in exist_keys:
+                        continue
+                    # add link from node outputs to group outputs
+                    self.add_link(node.outputs[key], self.outputs[new_key])
 
     def set_inputs(self, inputs: Dict[str, Any]):
         for name, input in inputs.items():
@@ -245,12 +299,6 @@ class NodeGraph:
         except importlib.metadata.PackageNotFoundError:
             return "unknown"
 
-    @property
-    def widget(self) -> NodeGraphWidget:
-        if self._widget is None:
-            self._widget = NodeGraphWidget(parent=self)
-        return self._widget
-
     def add_node(
         self,
         identifier: Union[str, Callable],
@@ -301,16 +349,6 @@ class NodeGraph:
         self._version += 1
         return link
 
-    def add_input(self, identifier: str, name: str, **kwargs) -> NodeSocket:
-        """Add an input socket to this node."""
-
-        input = self.inputs._new(identifier, name, **kwargs)
-        return input
-
-    def add_output(self, identifier: str, name: str, **kwargs) -> NodeSocket:
-        output = self.outputs._new(identifier, name, **kwargs)
-        return output
-
     def append_node(self, node: Node) -> None:
         """Appends a node to the node graph."""
         self.nodes._append(node)
@@ -346,6 +384,11 @@ class NodeGraph:
         Returns:
             Dict[str, Any]: The node graph data.
         """
+        # Capture the current ctx namespace shape into the graph spec before exporting.
+        if hasattr(self.graph_ctx, "inputs"):
+            ctx_snapshot = self.graph_ctx.inputs._to_spec()
+            self.spec = replace(self.spec, ctx=ctx_snapshot)
+
         metadata = self.get_metadata()
         nodes = self.export_nodes_to_dict(
             include_sockets=include_sockets, should_serialize=should_serialize
@@ -360,6 +403,7 @@ class NodeGraph:
             "action": self.action,
             "error": "",
             "metadata": metadata,
+            "spec": self.spec.to_dict(),
             "nodes": nodes,
             "links": links,
             "description": self.description,
@@ -367,23 +411,9 @@ class NodeGraph:
         return data
 
     def get_metadata(self) -> Dict[str, Any]:
-        """Export graph metadata including *live* graph-level IO specs.
-        We snapshot current shapes of graph inputs/outputs/ctx using `SocketSpec.from_namespace`.
-        """
-        self._inputs = SocketSpec.from_namespace(
-            self.graph_inputs.inputs, type_mapping=self.type_mapping
-        )
-        self._outputs = SocketSpec.from_namespace(
-            self.graph_outputs.inputs, type_mapping=self.type_mapping
-        )
-        self._ctx = SocketSpec.from_namespace(
-            self.graph_ctx.inputs, type_mapping=self.type_mapping
-        )
+        """Export graph metadata including *live* graph-level IO specs."""
         meta: Dict[str, Any] = {
             "graph_type": self.graph_type,
-            "inputs_spec": self._inputs.to_dict(),
-            "outputs_spec": self._outputs.to_dict(),
-            "ctx_spec": self._ctx.to_dict(),
         }
         # also save the parent class information
         meta["graph_class"] = {
@@ -455,7 +485,7 @@ class NodeGraph:
             )
 
     @classmethod
-    def from_dict(cls, ngdata: Dict[str, Any]) -> "NodeGraph":
+    def from_dict(cls, ngdata: Dict[str, Any]) -> NodeGraph:
         """Rebuilds a node graph from a dictionary.
 
         Args:
@@ -464,35 +494,18 @@ class NodeGraph:
         Returns:
             NodeGraph: The rebuilt node graph.
         """
-
-        md = ngdata.get("metadata", {}) or {}
-
-        ng: "NodeGraph" = cls(
+        spec = GraphSpec.from_dict(ngdata.get("spec", {}))
+        ng = cls(
             name=ngdata["name"],
             uuid=ngdata.get("uuid"),
+            inputs=spec.inputs,
+            outputs=spec.outputs,
+            ctx=spec.ctx,
             graph_type=ngdata["metadata"].get("graph_type", "NORMAL"),
-            state=ngdata.get("state", "CREATED"),
-            action=ngdata.get("action", "NONE"),
-            description=ngdata.get("description", ""),
-            init_graph_level_nodes=False,
         )
-        # built-in graph level nodes
-        inputs_spec = (
-            SocketSpec.from_dict(md["inputs_spec"], type_mapping=ng.type_mapping)
-            if "inputs_spec" in md
-            else None
-        )
-        outputs_spec = (
-            SocketSpec.from_dict(md["outputs_spec"], type_mapping=ng.type_mapping)
-            if "outputs_spec" in md
-            else None
-        )
-        ctx_spec = (
-            SocketSpec.from_dict(md["ctx_spec"], type_mapping=ng.type_mapping)
-            if "ctx_spec" in md
-            else None
-        )
-        ng._init_graph_level_nodes(inputs_spec, outputs_spec, ctx_spec)
+        ng.state = ngdata.get("state", "CREATED")
+        ng.action = ngdata.get("action", "NONE")
+        ng.description = ngdata.get("description", "")
 
         for ndata in ngdata["nodes"].values():
             ng.add_node_from_dict(ndata)
@@ -509,34 +522,13 @@ class NodeGraph:
         Returns:
             Node: The added node.
         """
-        import importlib
 
-        md = ndata.get("metadata", {}) or {}
         name = ndata["name"]
         if name in BUILTIN_NODES:
             self.nodes[name].update_from_dict(ndata)
             return self.nodes[name]
         else:
-            if md.get("is_dynamic", False):
-                node_class = ndata.get("metadata", {}).get("node_class", None)
-                node_class = getattr(
-                    importlib.import_module(node_class["module_path"]),
-                    node_class["callable_name"],
-                )
-                node = node_class.from_dict(ndata, graph=self)
-                self.nodes._append(node)
-            else:
-                identifier = md["identifier"]
-                node = self.add_node(
-                    identifier,
-                    name=name,
-                    uuid=ndata.pop("uuid", None),
-                    _metadata=ndata.get("metadata", None),
-                    _executor=ndata.get("executor", None),
-                    include_builtins=ndata["name"] in BUILTIN_NODES,
-                )
-                node.update_from_dict(ndata)
-        return node
+            return Node.from_dict(ndata, graph=self)
 
     @classmethod
     def from_yaml(
@@ -705,19 +697,6 @@ class NodeGraph:
     def __repr__(self) -> str:
         return f'NodeGraph(name="{self.name}", uuid="{self.uuid}")'
 
-    def _repr_mimebundle_(self, *args, **kwargs):
-        # if ipywdigets > 8.0.0, use _repr_mimebundle_ instead of _ipython_display_
-        self.widget.value = self.to_widget_value()
-        if hasattr(self.widget, "_repr_mimebundle_"):
-            return self.widget._repr_mimebundle_(*args, **kwargs)
-        else:
-            return self.widget._ipython_display_(*args, **kwargs)
-
-    def to_html(self, output: str = None, **kwargs):
-        """Write a standalone html file to visualize the graph."""
-        self.widget.value = self.to_widget_value()
-        return self.widget.to_html(output=output, **kwargs)
-
     def __enter__(self):
         from node_graph.manager import active_graph as _active_graph
 
