@@ -1,10 +1,12 @@
 from __future__ import annotations
+from uuid import uuid4
 from node_graph.collection import DependencyCollection
 from node_graph.property import NodeProperty
 from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union
 from node_graph.collection import get_item_class
-from dataclasses import MISSING, dataclass, field
+from dataclasses import MISSING, replace
 from node_graph.orm.mapping import type_mapping
+from node_graph.socket_meta import SocketMeta
 from node_graph.registry import EntryPointPool
 import wrapt
 
@@ -14,6 +16,17 @@ if TYPE_CHECKING:
     from node_graph.node_graph import NodeGraph
 
 
+def has_socket(data: dict):
+    """Check if the data contains a socket."""
+    for value in data.values():
+        if isinstance(value, BaseSocket):
+            return True
+        elif isinstance(value, dict):
+            return has_socket(value)
+
+    return False
+
+
 def op_add(x, y):
     return x + y
 
@@ -67,6 +80,73 @@ def op_ne(x, y):
     return x != y
 
 
+def _raise_illegal(sock, what: str, tips: list[str]):
+    from .errors import GraphDeferredIllegalOperationError
+
+    node = getattr(sock, "_node", None)
+    node_name = getattr(node, "name", None) or "<unknown-node>"
+    socket_name = (
+        getattr(sock, "_name", None) or getattr(sock, "name", None) or "<socket>"
+    )
+
+    common = [
+        "General guidance:",
+        "  • Wrap logic in a nested @node.graph.",
+        "  • Or use the WorkGraph If zone for branching on predicates.",
+        "  • Or for loops, use the While zone or Map zone.",
+    ]
+
+    msg = (
+        f"Illegal operation on a future value (Socket): {what}\n"
+        f"Socket: {socket_name} of node '{node_name}'"
+    )
+    msg += "\n\nFix:\n" + "\n".join(tips + [""] + common)
+    raise GraphDeferredIllegalOperationError(msg)
+
+
+def _tip_cast(kind):  # numeric/bytes/path-like casts
+    return [
+        f"Avoid {kind} on futures. Compute the value inside the graph, then cast afterwards.",
+        "If you need a cast during execution, use a dedicated cast node.",
+    ]
+
+
+def _tip_iter():  # iteration/len/container-ish
+    return [
+        "You tried to iterate or take len()/index a future.",
+        "Use @node.graph to build logic that needs to iterate over values.",
+    ]
+
+
+def _tip_bool():
+    return [
+        "You used a future in a boolean context (if/while/assert/and/or).",
+        "Wrap logic in a nested @node.graph.",
+    ]
+
+
+def _tip_numpy():
+    return [
+        "NumPy tried to coerce or operate on a future.",
+        "Use built-in operator sockets (+, -, *, <, …) to build predicates/expressions,",
+        "or use a dedicated numpy/ufunc node.",
+    ]
+
+
+def _tip_ctxmgr():
+    return [
+        "You tried to use a future as a context manager.",
+        "Wrap side effects in a graph node or zone instead.",
+    ]
+
+
+def _tip_indexing():
+    return [
+        "You tried to subscript a future (obj[idx]).",
+        "Index inside the graph (node/zone) where the value is concrete.",
+    ]
+
+
 class OperatorSocketMixin:
     @property
     def _decorator(self):
@@ -181,6 +261,134 @@ class OperatorSocketMixin:
             self._waiting_on.add(other)
         return other
 
+    # Truthiness / boolean contexts
+    def __bool__(self):
+        _raise_illegal(self, "boolean evaluation", _tip_bool())
+
+    # Numeric casts & indices
+    def __int__(self):
+        _raise_illegal(self, "int() cast", _tip_cast("int()"))
+
+    def __float__(self):
+        _raise_illegal(self, "float() cast", _tip_cast("float()"))
+
+    def __complex__(self):
+        _raise_illegal(self, "complex() cast", _tip_cast("complex()"))
+
+    def __index__(self):
+        _raise_illegal(self, "use as an index (__index__)", _tip_cast("indexing"))
+
+    def __round__(self, *a, **k):
+        _raise_illegal(self, "round()", _tip_cast("round()"))
+
+    def __trunc__(self):
+        _raise_illegal(self, "trunc()", _tip_cast("trunc()"))
+
+    def __floor__(self):
+        _raise_illegal(self, "math.floor()", _tip_cast("floor()"))
+
+    def __ceil__(self):
+        _raise_illegal(self, "math.ceil()", _tip_cast("ceil()"))
+
+    # Sequence / mapping / container protocols
+    def __len__(self):
+        _raise_illegal(self, "len()", _tip_iter())
+
+    def __iter__(self):
+        _raise_illegal(self, "iteration", _tip_iter())
+
+    def __reversed__(self):
+        _raise_illegal(self, "reversed()", _tip_iter())
+
+    def __contains__(self, _):
+        _raise_illegal(self, "membership test (x in socket)", _tip_iter())
+
+    def __getitem__(self, _):
+        _raise_illegal(self, "subscript access (socket[idx])", _tip_indexing())
+
+    def __setitem__(self, *_):
+        _raise_illegal(self, "item assignment (socket[idx] = ...)", _tip_indexing())
+
+    def __delitem__(self, *_):
+        _raise_illegal(self, "item deletion (del socket[idx])", _tip_indexing())
+
+    # Bitwise / logical operators that people misuse for predicates
+    def __and__(self, _):
+        _raise_illegal(self, "bitwise and (&) on futures", _tip_bool())
+
+    def __or__(self, _):
+        _raise_illegal(self, "bitwise or (|) on futures", _tip_bool())
+
+    def __xor__(self, _):
+        _raise_illegal(self, "bitwise xor (^) on futures", _tip_bool())
+
+    def __invert__(self):
+        _raise_illegal(self, "bitwise not (~) on futures", _tip_bool())
+
+    def __matmul__(self, _):
+        _raise_illegal(self, "matrix multiply (@)", _tip_numpy())
+
+    # Hashing / dict keys / set members
+    def __hash__(self):
+        _raise_illegal(
+            self,
+            "hashing (use as dict/set key)",
+            ["Futures are not stable keys. Resolve to a concrete value first."],
+        )
+
+    # Function-like / context-manager / async
+    def __enter__(self):
+        _raise_illegal(
+            self,
+            "context-manager enter (__enter__)",
+            ["Use Socket as a context manager is not supported."],
+        )
+
+    def __exit__(self, *a):
+        _raise_illegal(
+            self,
+            "context-manager exit (__exit__)",
+            ["Use Socket as a context manager is not supported."],
+        )
+
+    def __await__(self):
+        _raise_illegal(
+            self,
+            "await on a future (__await__)",
+            ["Awaiting is not supported; use @node.graph to build logic instead."],
+        )
+
+    def __aiter__(self):
+        _raise_illegal(
+            self,
+            "async iteration (__aiter__)",
+            [
+                "Async iteration is not supported; use @node.graph to build logic instead."
+            ],
+        )
+
+    def __anext__(self):
+        _raise_illegal(
+            self,
+            "async next (__anext__)",
+            [
+                "Async iteration is not supported; use @node.graph to build logic instead."
+            ],
+        )
+
+    # NumPy interoperability guards
+    # Prevent silent coercion to ndarray
+    def __array__(self, *a, **k):
+        _raise_illegal(self, "NumPy array coercion (__array__)", _tip_numpy())
+
+    # Intercept ufuncs like np.add, np.sin, etc.
+    def __array_ufunc__(self, *a, **k):
+        _raise_illegal(self, "NumPy ufunc on future (__array_ufunc__)", _tip_numpy())
+
+    # Intercept high-level NumPy functions
+    def __array_function__(self, *a, **k):
+        _raise_illegal(self, "NumPy high-level op (__array_function__)", _tip_numpy())
+
 
 class WaitingOn:
     """
@@ -208,84 +416,31 @@ class WaitingOn:
             print(f"Link {link_name} already exists, skipping creation.")
 
 
-@dataclass()
-class SocketMetadata:
-    """A *typed* container for additional socket information.
+def _normalize_meta(raw: Union[SocketMeta, Dict[str, Any], None]) -> SocketMeta:
+    if raw is None:
+        return SocketMeta()
+    if isinstance(raw, SocketMeta):
+        return SocketMeta.from_dict(raw.to_dict())
+    if isinstance(raw, dict):
+        meta = SocketMeta.from_dict(raw)
+        if meta.required is None:
+            meta.required = False
+        if meta.socket_type is None:
+            meta.socket_type = "INPUT"
+        if meta.arg_type is None:
+            meta.arg_type = "kwargs"
+        return meta
+    raise TypeError(f"metadata must be dict | SocketMeta | None – got {type(raw)!r}")
 
-    Parameters
-    ----------
-    dynamic
-        Whether the socket collection is *dynamic* - i.e. it may grow
-        automatically when assigning unknown keys.
-    builtin_socket
-        Marks sockets that are intrinsic to the framework (e.g. ``_wait`` or
-        ``outputs``) so that user code can filter / style them differently.
-    extras
-        Free form mapping for user extensions.  Any key not matching one of the
-        reserved field names ends up in here when converting from an untyped
-        ``dict``.
-    """
 
-    dynamic: bool = False
-    sub_socket_default_link_limit: int = 1
-    required: bool = False
-    builtin_socket: bool = False
-    function_socket: bool = False
-    socket_type: str = "INPUT"
-    arg_type: str = "kwargs"
-    extras: Dict[str, Any] = field(default_factory=dict)
-
-    def to_dict(self) -> Dict[str, Any]:
-        """Return a *plain* dict suitable for JSON serialisation."""
-
-        data = {
-            "dynamic": self.dynamic,
-            "required": self.required,
-            "builtin_socket": self.builtin_socket,
-            "function_socket": self.function_socket,
-            "socket_type": self.socket_type,
-            "arg_type": self.arg_type,
-        }
-        if self.extras:
-            # Make sure extras are a dict
-            data["extras"] = dict(self.extras)
-
-        # Do not bloat output with empty *extras*
-        if not data.get("extras"):
-            data.pop("extras", None)
-        return data
-
-    @classmethod
-    def from_raw(
-        cls, raw: Union["SocketMetadata", Dict[str, Any], None]
-    ) -> "SocketMetadata":
-        """Normalise *raw* user input into a :class:`SocketMetadata` instance."""
-
-        if raw is None:
-            return cls()
-        if isinstance(raw, cls):
-            return raw
-        if isinstance(raw, dict):
-            # Extract known keys and forward unknown ones into *extras*
-            known_keys = {
-                "dynamic",
-                "builtin_socket",
-                "function_socket",
-                "socket_type",
-                "arg_type",
-                "required",
-                "sub_socket_default_link_limit",
-                "extras",
-            }
-            known = {k: v for k, v in raw.items() if k in known_keys}
-            known.setdefault("extras", {})
-            known["extras"].update(
-                {k: v for k, v in raw.items() if k not in known_keys}
-            )
-            return cls(**known)
-        raise TypeError(
-            f"metadata must be dict | SocketMetadata | None – got {type(raw)!r}"
-        )
+_RUNTIME_EXTRA_KEYS = {
+    "identifier",
+    "sockets",
+    "dynamic",
+    "item",
+    "builtin_socket",
+    "function_socket",
+}
 
 
 class TaggedValue(wrapt.ObjectProxy):
@@ -293,12 +448,17 @@ class TaggedValue(wrapt.ObjectProxy):
         super().__init__(wrapped)
 
         self._self_socket = socket
+        self._self_uuid = str(uuid4())
 
     # Provide clean access via `proxy._socket` instead of `proxy._self_socket`
     @property
     def _socket(self):
         return self._self_socket
 
+    @property
+    def _uuid(self):
+        return self._self_uuid
+
     @_socket.setter
     def _socket(self, value):
         self._self_socket = value
@@ -353,7 +513,7 @@ class BaseSocket:
         parent: Optional["NodeSocketNamespace"] = None,
         graph: Optional["NodeGraph"] = None,
         link_limit: int = 1,
-        metadata: Union[SocketMetadata, Dict[str, Any], None] = None,
+        metadata: Union[SocketMeta, Dict[str, Any], None] = None,
         **kwargs: Any,
     ) -> None:
         """Initialize an instance of NodeSocket.
@@ -373,7 +533,7 @@ class BaseSocket:
         self._graph = graph
         self._links = []
         self._link_limit = link_limit
-        self._metadata: SocketMetadata = SocketMetadata.from_raw(metadata)
+        self._metadata: SocketMeta = _normalize_meta(metadata)
         self._waiting_on = WaitingOn(node=self._node, graph=self._graph)
 
     @property
@@ -505,7 +665,7 @@ class NodeSocket(BaseSocket, OperatorSocketMixin):
             ):
                 value = value._outputs
             self._node.graph.add_link(value, self)
-        elif isinstance(value, TaggedValue):
+        elif isinstance(value, TaggedValue) and value._socket is not None:
             self._node.graph.add_link(value._socket, self)
         elif self.property:
             self.property.value = value
@@ -523,6 +683,42 @@ class NodeSocket(BaseSocket, OperatorSocketMixin):
             data["property"] = None
         return data
 
+    def _to_spec(self) -> "SocketSpec":
+        """
+        Create a SocketSpec describing the current runtime state of this socket.
+        """
+        from copy import deepcopy
+        from node_graph.socket_spec import SocketSpec, SocketMeta
+
+        runtime_meta = self._metadata
+        extras = {
+            k: v
+            for k, v in (runtime_meta.extras or {}).items()
+            if k not in _RUNTIME_EXTRA_KEYS
+        }
+
+        meta = SocketMeta(
+            help=runtime_meta.help,
+            required=runtime_meta.required,
+            call_role=runtime_meta.call_role,
+            is_metadata=runtime_meta.is_metadata,
+            dynamic=False,
+            child_default_link_limit=runtime_meta.child_default_link_limit,
+            extras=extras,
+        )
+        spec_kwargs: Dict[str, Any] = {
+            "identifier": self._identifier,
+            "item": None,
+            "fields": {},
+            "link_limit": self._link_limit,
+            "meta": meta,
+        }
+
+        if "default" in extras:
+            spec_kwargs["default"] = deepcopy(extras["default"])
+
+        return SocketSpec(**spec_kwargs)
+
     @classmethod
     def _from_dict(cls, data: Dict[str, Any]) -> None:
         # Create a new instance of this class
@@ -579,6 +775,32 @@ def check_identifier_name(identifier: str, pool: dict) -> None:
         raise ValueError(msg)
 
 
+def _raise_namespace_assignment_error(
+    *,
+    target_ns: "NodeSocketNamespace",
+    incoming_desc: str,
+    reason: str,
+    fixes: list[str],
+) -> None:
+    """Raise a ValueError guiding users when setting/linking values into a namespace."""
+    where = getattr(target_ns, "_full_name_with_node", "<namespace>")
+    msg = [
+        f"Invalid assignment into namespace socket: {where}",
+        "",
+        "What happened:",
+        f"  • {reason}",
+        f"  • Incoming value: {incoming_desc}",
+        "",
+        "Why this matters:",
+        "  • Without a namespace-typed graph input, the whole dict is treated as *one value*,",
+        "    so its inner keys are NOT wired to the child sockets. You'd end up with missing links.",
+        "",
+        "How to fix:",
+        *[f"  • {line}" for line in fixes],
+    ]
+    raise ValueError("\n".join(msg))
+
+
 class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
     """A NodeSocket that also acts as a namespace (collection) of other sockets."""
 
@@ -603,7 +825,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         node: Optional["Node"] = None,
         parent: Optional["NodeSocket"] = None,
         link_limit: int = 1000000,
-        metadata: Union[SocketMetadata, Dict[str, Any], None] = None,
+        metadata: Union[SocketMeta, Dict[str, Any], None] = None,
         sockets: Optional[Dict[str, object]] = None,
         pool: Optional[object] = None,
         entry_point: Optional[str] = None,
@@ -661,9 +883,11 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         try:
             return self._sockets[name]
         except KeyError:
+            avail = ", ".join(self._sockets.keys()) or "<none>"
             raise AttributeError(
-                f"{self.__class__.__name__}: {self._full_name_with_node} has no attribute '{name}'. "
-                f"Available sub-sockets are: {', '.join(self._sockets.keys())}"
+                f"{self.__class__.__name__}: '{self._full_name_with_node}' has no sub-socket '{name}'.\n"
+                f"Available: {avail}\n"
+                f"Tip: If '{name}' should exist, add it to the SocketSpec (or make this namespace dynamic)."
             )
 
     def __setattr__(self, name: str, value: Any) -> None:
@@ -677,9 +901,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             object.__setattr__(self, name, value)
             return
 
-        self._set_socket_value(
-            {name: value}, link_limit=self._metadata.sub_socket_default_link_limit
-        )
+        self._set_socket_value({name: value})
 
     def __setitem__(self, key: str | int, value: Any) -> None:
         """
@@ -714,6 +936,10 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         identifier = identifier or self._SocketPool["any"]
         check_identifier_name(identifier, self._SocketPool)
 
+        meta_payload = dict(metadata or {})
+        if "socket_type" not in meta_payload and self._metadata.socket_type:
+            meta_payload["socket_type"] = self._metadata.socket_type
+
         _names = name.split(".", 1)
         if len(_names) > 1:
             namespace = _names[0]
@@ -726,7 +952,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                         namespace,
                         metadata={
                             "dynamic": True,
-                            "sub_socket_default_link_limit": self._metadata.sub_socket_default_link_limit,
+                            "child_default_link_limit": self._metadata.child_default_link_limit,
                         },
                     )
                 else:
@@ -736,21 +962,19 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             return self[namespace]._new(
                 identifier,
                 _names[1],
-                link_limit=self._metadata.sub_socket_default_link_limit,
-                metadata=metadata,
+                link_limit=self._metadata.child_default_link_limit,
+                metadata=meta_payload,
             )
         else:
             ItemClass = get_item_class(identifier, self._SocketPool)
             kwargs.pop("graph", None)
-            kwargs.setdefault(
-                "link_limit", self._metadata.sub_socket_default_link_limit
-            )
+            kwargs.setdefault("link_limit", self._metadata.child_default_link_limit)
             item = ItemClass(
                 name,
                 node=self._node,
                 parent=self,
                 graph=self._graph,
-                metadata=metadata,
+                metadata=meta_payload,
                 pool=self._SocketPool,
                 **kwargs,
             )
@@ -780,11 +1004,101 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                         data[name] = item.value
         return data
 
+    def _to_spec(self) -> "SocketSpec":
+        """
+        Materialize the current namespace into a SocketSpec snapshot.
+        """
+        from node_graph.socket_spec import SocketSpec, SocketMeta
+        from copy import deepcopy
+
+        runtime_meta = self._metadata
+        extras = {
+            k: v
+            for k, v in (runtime_meta.extras or {}).items()
+            if k not in _RUNTIME_EXTRA_KEYS
+        }
+
+        meta = SocketMeta(
+            help=runtime_meta.help,
+            required=runtime_meta.required,
+            call_role=runtime_meta.call_role,
+            is_metadata=runtime_meta.is_metadata,
+            dynamic=self._metadata.dynamic,
+            child_default_link_limit=runtime_meta.child_default_link_limit,
+            extras=extras,
+        )
+
+        fields: Dict[str, "SocketSpec"] = {}
+        for name, child in self._sockets.items():
+            if child._metadata.extras.get("builtin_socket"):
+                continue
+            if hasattr(child, "_to_spec"):
+                fields[name] = child._to_spec()
+
+        item_spec = None
+        if "item" in extras:
+            item_snapshot = extras["item"]
+            if isinstance(item_snapshot, dict):
+                item_spec = self._spec_from_shape_snapshot(item_snapshot)
+            else:
+                item_spec = None
+
+        spec = SocketSpec(
+            identifier=self._identifier,
+            item=item_spec,
+            fields=fields,
+            link_limit=self._link_limit,
+            meta=meta,
+        )
+
+        if "default" in extras:
+            spec = replace(spec, default=deepcopy(extras["default"]))
+
+        return spec
+
+    @staticmethod
+    def _spec_from_shape_snapshot(snapshot: Dict[str, Any]) -> "SocketSpec":
+        """
+        Rebuild a SocketSpec from a minimal shape snapshot stored in metadata extras.
+        """
+        from copy import deepcopy
+        from node_graph.socket_spec import SocketSpec, SocketMeta
+
+        identifier = snapshot.get("identifier", "node_graph.any")
+
+        fields = {
+            name: NodeSocketNamespace._spec_from_shape_snapshot(child_snapshot)
+            for name, child_snapshot in snapshot.get("sockets", {}).items()
+        }
+
+        item_snapshot = snapshot.get("item")
+        item_spec = (
+            NodeSocketNamespace._spec_from_shape_snapshot(item_snapshot)
+            if isinstance(item_snapshot, dict)
+            else None
+        )
+
+        meta = SocketMeta(dynamic=bool(snapshot.get("dynamic", False)))
+
+        spec_kwargs: Dict[str, Any] = {
+            "identifier": identifier,
+            "fields": fields,
+            "item": item_spec,
+            "meta": meta,
+        }
+
+        if "default" in snapshot:
+            spec_kwargs["default"] = deepcopy(snapshot["default"])
+        if "link_limit" in snapshot:
+            spec_kwargs["link_limit"] = snapshot["link_limit"]
+
+        return SocketSpec(**spec_kwargs)
+
     @_value.setter
     def _value(self, value: Dict[str, Any]) -> None:
         self._set_socket_value(value)
 
-    def _set_socket_value(self, value: Dict[str, Any] | NodeSocket, **kwargs) -> None:
+    def _set_socket_value(self, value: Dict[str, Any] | NodeSocket) -> None:
         """Set value(s) into this namespace.
 
         Supports:
@@ -796,6 +1110,8 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         - Missing immediate children can be created only if *this* namespace is dynamic.
         - Intermediate dotted segments are created as namespaces (dynamic=True) when needed.
         """
+        from node_graph.socket_spec import SocketSpec
+
         if value is None:
             return
 
@@ -804,9 +1120,36 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             self._node.graph.add_link(value, self)
             return
 
+        if isinstance(value, TaggedValue):
+            src = getattr(
+                getattr(value, "_socket", None),
+                "_full_name_with_node",
+                "<unknown-socket>",
+            )
+            _raise_namespace_assignment_error(
+                target_ns=self,
+                incoming_desc=f"TaggedValue(dict) from {src}",
+                reason="A TaggedValue wrapping a dict is being assigned to a *namespace*.",
+                fixes=[
+                    "Annotate the graph-level parameter as a namespace, e.g.:",
+                    f"    def your_graph({self._name}: Annotated[dict, namespace({list(self._sockets.keys())[0]}=, ...)]):",
+                    "",
+                    "Or reuse the node’s own input spec, e.g.:",
+                    f"    def your_graph({self._name}: Annotated[dict, add_multiply.inputs.data]):",
+                    "        add_multiply(data=data)",
+                    "",
+                ],
+            )
+
         if not isinstance(value, dict):
-            raise ValueError(
-                f"Invalid value type for socket {self._name}: {value}, expected dict or Socket."
+            _raise_namespace_assignment_error(
+                target_ns=self,
+                incoming_desc=type(value).__name__,
+                reason="This is a namespace socket and expects a dict (or a Socket link).",
+                fixes=[
+                    "Provide a dict with keys matching the namespace fields;",
+                    "or link another socket/namespace; or declare the graph input as a namespace.",
+                ],
             )
 
         for key, val in value.items():
@@ -817,9 +1160,15 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                 # Ensure the immediate child exists (create if allowed)
                 if head not in self._sockets:
                     if not self._metadata.dynamic:
-                        raise ValueError(
-                            f"Socket: {head} does not exist in non-dynamic namespace '{self._name}'. "
-                            f"This namespace belongs to: {self._node.name}"
+                        _raise_namespace_assignment_error(
+                            target_ns=self,
+                            incoming_desc=f"key '{head}'",
+                            reason=f"Field '{head}' is not defined and this namespace is not dynamic.",
+                            fixes=[
+                                "Define the field in the socket spec (preferred); or",
+                                "mark this namespace dynamic if it must accept arbitrary keys;",
+                                "or correct the key path being assigned.",
+                            ],
                         )
                     # We are going to descend (tail exists), so create a namespace
                     self._new(
@@ -827,58 +1176,81 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                         head,
                         metadata={
                             "dynamic": True,
-                            "sub_socket_default_link_limit": self._metadata.sub_socket_default_link_limit,
+                            "child_default_link_limit": self._metadata.child_default_link_limit,
                         },
-                        **kwargs,
                     )
 
                 child = self._sockets[head]
                 if not isinstance(child, NodeSocketNamespace):
-                    raise ValueError(
-                        f"Cannot set nested key '{key}' under non-namespace socket '{head}'."
+                    _raise_namespace_assignment_error(
+                        target_ns=self,
+                        incoming_desc=f"nested key '{key}' under leaf '{head}'",
+                        reason=f"'{head}' is a leaf socket, but you attempted to assign nested data below it.",
+                        fixes=[
+                            "Use a namespace socket for hierarchical data; update your SocketSpec accordingly; or",
+                            "flatten your assignment to target a leaf socket directly.",
+                        ],
                     )
 
                 # Recurse into the child namespace with the remaining tail
-                child._set_socket_value({tail: val}, **kwargs)
+                child._set_socket_value({tail: val})
                 continue  # next key
 
             # Non-dotted key path (single-segment)
             if key not in self._sockets:
                 if not self._metadata.dynamic:
-                    raise ValueError(
-                        f"Socket: {key} does not exist in a non-dynamic namespace socket: {self._name}. "
-                        f"This namespace belongs to: {self._node.name}"
+                    _raise_namespace_assignment_error(
+                        target_ns=self,
+                        incoming_desc=f"key '{key}'",
+                        reason=f"Field '{key}' is not defined and this namespace is not dynamic.",
+                        fixes=[
+                            "Add the field to the namespace spec; or",
+                            "make the namespace dynamic if it should grow automatically.",
+                        ],
                     )
 
-                # Decide whether to create a namespace or a leaf based on the value type
-                if isinstance(val, (dict, NodeSocketNamespace)):
-                    # create child namespace (dynamic)
-                    self._new(
-                        self._SocketPool["namespace"],
-                        key,
-                        metadata={
+                # Create a leaf or namespace based on the dynamic item type
+                extras = self._metadata.extras or {}
+                item_snapshot = extras.get("item") if isinstance(extras, dict) else None
+                if item_snapshot is None:
+                    if isinstance(val, (dict, NodeSocketNamespace)):
+                        item_snapshot = {
+                            "identifier": self._type_mapping["namespace"],
                             "dynamic": True,
-                            "sub_socket_default_link_limit": self._metadata.sub_socket_default_link_limit,
-                        },
-                        **kwargs,
-                    )
-                else:
-                    # create a leaf socket that can accept "any"
-                    self._new(self._SocketPool["any"], key, **kwargs)
+                        }
+                    else:
+                        item_snapshot = {"identifier": self._type_mapping["default"]}
+                item_spec = (
+                    SocketSpec.from_dict(item_snapshot)
+                    if isinstance(item_snapshot, dict)
+                    else None
+                )
+                self._append_from_spec(
+                    self,
+                    key,
+                    item_spec,
+                    node=self._node,
+                    graph=self._graph,
+                    role="input",
+                )
 
             # Now we’re guaranteed the key exists; delegate appropriately
             target = self._sockets[key]
             if isinstance(target, NodeSocketNamespace):
                 # If incoming val is a dict, recurse. If it’s a socket, link to the namespace.
                 if isinstance(val, dict):
-                    target._set_socket_value(val, **kwargs)
+                    target._set_socket_value(val)
                 elif isinstance(val, BaseSocket):
                     self._node.graph.add_link(val, target)
                 else:
                     # Treat setting a leaf value into a namespace as error for clarity
-                    raise ValueError(
-                        f"Cannot set a non-dict value on namespace '{target._full_name_with_node}'. "
-                        f"Got: {type(val).__name__}"
+                    _raise_namespace_assignment_error(
+                        target_ns=target,
+                        incoming_desc=type(val).__name__,
+                        reason="A namespace expects a mapping of fields, but a non-dict value was provided.",
+                        fixes=[
+                            "Provide a dict like {'x': 1, 'y': 2} that matches the namespace shape.",
+                        ],
                     )
             else:
                 # Leaf socket: forward to its own setter (which handles linking or value assignment)
@@ -958,8 +1330,9 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             node=node,
             parent=parent,
             graph=graph,
+            link_limit=spec.link_limit or 1,
             metadata=ns_meta,
-            pool=pool or (node.SocketPool if node else None),
+            pool=pool or (node._REGISTRY.socket_pool if node else None),
         )
 
         # materialize fixed fields
@@ -978,6 +1351,30 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         graph,
         role: str,
     ) -> None:
+        if "." in name:
+            head, tail = name.split(".", 1)
+            existing = parent_ns._sockets.get(head)
+
+            if existing is None:
+                raise ValueError(
+                    f"Cannot assign nested field '{tail}' under missing socket '{head}'."
+                )
+
+            if not isinstance(existing, NodeSocketNamespace):
+                raise TypeError(
+                    f"Cannot assign nested field '{tail}' under non-namespace socket '{head}'."
+                )
+
+            cls._append_from_spec(
+                existing,
+                tail,
+                spec,
+                node=node,
+                graph=graph,
+                role=role,
+            )
+            return
+
         from copy import deepcopy
         from node_graph.materialize import runtime_meta_from_spec
 
@@ -1022,7 +1419,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                 graph=graph,
                 metadata=leaf_meta,
                 property=prop,
-                link_limit=parent_ns._metadata.sub_socket_default_link_limit,
+                link_limit=parent_ns._metadata.child_default_link_limit,
             )
             parent_ns._append(sock)
 
